package io.gitlab.jfronny.inceptum.gson;

import com.google.gson.*;
import com.google.gson.reflect.TypeToken;
import io.gitlab.jfronny.inceptum.model.inceptum.source.CurseforgeModSource;
import io.gitlab.jfronny.inceptum.model.inceptum.source.DirectModSource;
import io.gitlab.jfronny.inceptum.model.inceptum.source.ModSource;
import io.gitlab.jfronny.inceptum.model.inceptum.source.ModrinthModSource;

import java.io.IOException;
import java.lang.reflect.Type;
import java.util.Set;

public class ModSourceTypeAdapter implements JsonSerializer<ModSource>, JsonDeserializer<ModSource> {
    private static final Type modSourceSetType = new TypeToken<Set<ModSource>>() {}.getType();
    @Override
    public ModSource deserialize(JsonElement json, Type typeOfT, JsonDeserializationContext context) throws JsonParseException {
        if (!(json instanceof JsonObject jo))
            throw new JsonParseException("Expected ModSource to be an object");
        if (!jo.has("type"))
            throw new JsonParseException("Expected ModSource to contain a type");
        return switch (jo.get("type").getAsString()) {
            case "modrinth" -> {
                if (!jo.has("id"))
                    throw new JsonParseException("Expected ModrinthModSource to contain a version ID");
                try {
                    yield new ModrinthModSource(jo.get("id").getAsString());
                } catch (IOException e) {
                    throw new JsonParseException("Could not fetch Modrinth source", e);
                }
            }
            case "curseforge" -> {
                if (!jo.has("projectId"))
                    throw new JsonParseException("Expected a projectId in this curseforge project");
                if (!jo.has("fileId"))
                    throw new JsonParseException("Expected a fileId in this curseforge project");
                try {
                    yield new CurseforgeModSource(jo.get("projectId").getAsInt(), jo.get("fileId").getAsInt());
                } catch (IOException e) {
                    throw new JsonParseException("Could not fetch Curseforge source", e);
                }
            }
            case "direct" -> {
                if (!jo.has("fileName"))
                    throw new JsonParseException("Expected direct download to have a fileName");
                if (!jo.has("url"))
                    throw new JsonParseException("Expected direct download to have a url");
                if (jo.has("dependencies"))
                    yield new DirectModSource(jo.get("fileName").getAsString(), jo.get("url").getAsString(), context.deserialize(jo.get("dependencies"), modSourceSetType));
                else
                    yield new DirectModSource(jo.get("fileName").getAsString(), jo.get("url").getAsString(), Set.of());
            }
            default -> throw new JsonParseException("Unexpected ModSource type: " + jo.get("type").getAsString());
        };
    }

    @Override
    public JsonElement serialize(ModSource src, Type typeOfSrc, JsonSerializationContext context) {
        JsonObject jo = new JsonObject();
        if (src instanceof ModrinthModSource mo) {
            jo.add("type", new JsonPrimitive("modrinth"));
            jo.add("id", new JsonPrimitive(mo.getVersionId()));
        }
        else if (src instanceof DirectModSource di) {
            jo.add("type", new JsonPrimitive("direct"));
            jo.add("fileName", new JsonPrimitive(di.fileName()));
            jo.add("url", new JsonPrimitive(di.url()));
            jo.add("dependencies", context.serialize(di.dependencies()));
        }
        else if (src instanceof CurseforgeModSource cu) {
            jo.add("type", new JsonPrimitive("curseforge"));
            jo.add("projectId", new JsonPrimitive(cu.getProjectId()));
            jo.add("fileId", new JsonPrimitive(cu.getFileId()));
        }
        else throw new RuntimeException("ModSources with the type " + src.getClass() + " are not supported");
        return jo;
    }
}
