package io.gitlab.jfronny.inceptum.model.inceptum.source;

import io.gitlab.jfronny.inceptum.model.modrinth.ModrinthMod;
import io.gitlab.jfronny.inceptum.model.modrinth.ModrinthVersion;
import io.gitlab.jfronny.inceptum.util.HashUtils;
import io.gitlab.jfronny.inceptum.util.Utils;
import io.gitlab.jfronny.inceptum.util.api.ModrinthApi;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.HashSet;
import java.util.Optional;
import java.util.Set;

public final class ModrinthModSource implements ModSource {
    private final String versionId;
    private final ModrinthVersion current;
    private final ModrinthMod mod;

    public ModrinthModSource(String versionId) throws IOException {
        this.versionId = versionId;
        this.current = ModrinthApi.getVersion(versionId);
        this.mod = ModrinthApi.getMod(getModId());
    }

    @Override
    public ModDownload download() throws IOException {
        ModrinthVersion.File file = current.files.get(0);
        Path path = getJarPath();
        Utils.downloadFile(file.url, file.hashes.sha1, path);
        return new ModDownload(file.hashes.sha1, HashUtils.murmur2(Files.readAllBytes(path)), path);
    }

    @Override
    public Set<ModSource> getDependencies() throws IOException {
        Set<ModSource> deps = new HashSet<>();
        for (ModrinthVersion.Dependency dependency : current.dependencies) {
            //TODO show optional dependencies
            if (dependency.dependency_type == ModrinthVersion.Dependency.DependencyType.required)
                deps.add(new ModrinthModSource(dependency.version_id));
        }
        return deps;
    }

    @Override
    public Optional<ModSource> getUpdate(String gameVersion) throws IOException {
        ModrinthVersion stable = null;
        ModrinthVersion beta = null;
        ModrinthVersion latest = null;
        for (ModrinthVersion version : ModrinthApi.getVersions(getModId())) {
            if (version.game_versions.contains(gameVersion) && version.loaders.contains("fabric")) {
                if (latest == null) latest = version;
                if (version.version_type == ModrinthVersion.VersionType.beta || version.version_type == ModrinthVersion.VersionType.release) {
                    beta = version;
                }
                if (version.version_type == ModrinthVersion.VersionType.release) {
                    stable = version;
                }
            }
        }
        ModrinthVersion next = switch (current.version_type) {
            case alpha -> latest;
            case beta -> beta;
            case release -> stable;
        };
        if (next == null) return Optional.empty();
        if (next.version_number.equals(current.version_number)) return Optional.empty();
        return Optional.of(new ModrinthModSource(next.id));
    }

    @Override
    public String getVersion() {
        return current.version_number;
    }

    @Override
    public String getName() {
        return "modrinth/" + (mod.slug == null ? mod.id : mod.slug) + '/' + current.version_number;
    }

    @Override
    public String getFileName() {
        return current.files.get(0).filename;
    }

    @Override
    public boolean equals(ModSource other) {
        return other instanceof ModrinthModSource ms && ms.getModId().equals(getModId());
    }

    public String getVersionId() {
        return versionId;
    }

    public String getModId() {
        return current.mod_id;
    }
}
