package io.gitlab.jfronny.inceptum.util;

import io.gitlab.jfronny.inceptum.Inceptum;
import io.gitlab.jfronny.inceptum.model.fabric.FabricModJson;
import io.gitlab.jfronny.inceptum.model.inceptum.InstanceMeta;
import io.gitlab.jfronny.inceptum.model.inceptum.ModDescription;
import io.gitlab.jfronny.inceptum.model.inceptum.source.ModSource;

import java.io.Closeable;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URISyntaxException;
import java.nio.file.FileSystem;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.*;

public class ModsDirScanner implements Closeable {
    private final Map<Path, IWModDescription> descriptions = new HashMap<>();
    private final Thread th;
    private final Path modsDir;
    private final InstanceMeta instance;
    private boolean disposed = false;

    public ModsDirScanner(Path modsDir, InstanceMeta instance) {
        this.modsDir = modsDir;
        this.instance = instance;
        this.th = new Thread(this::scanTaskInternal);
    }

    public void start() {
        th.start();
    }

    public String getGameVersion() {
        return instance.getMinecraftVersion();
    }

    public Set<IWModDescription> getMods() throws IOException {
        Set<IWModDescription> mods = new TreeSet<>();
        if (Files.isDirectory(modsDir)) {
            for (Path path : Utils.ls(modsDir)) {
                mods.add(get(path));
            }
        }
        return mods;
    }

    public IWModDescription get(Path path) {
        if (descriptions.containsKey(path))
            return descriptions.get(path);
        else {
            // not yet scanned
            return new IWModDescription(path); //TODO figure out why this causes blocking
        }
    }

    private void scanTaskInternal() {
        while (!disposed) {
            try {
                if (!Files.isDirectory(modsDir)) {
                    Thread.sleep(5000);
                    continue;
                }
                for (Path mods : Utils.ls(modsDir)) {
                    if (Files.isDirectory(mods)) {
                        descriptions.put(mods, new IWModDescription(mods));
                    } else {
                        if (mods.toString().endsWith(".jar") || mods.toString().endsWith(".jar.disabled")) {
                            Optional<Path> imod = Optional.of(mods.getParent().resolve(mods.getFileName() + ".imod"));
                            // mod description
                            if (!Files.exists(imod.get())) {
                                Utils.writeObject(imod.get(), ModDescription.of(mods));
                            }
                            ModDescription md = Utils.loadObject(imod.get(), ModDescription.class);
                            evaluateSources(md, mods);
                            descriptions.put(imod.get(), new IWModDescription(mods, Optional.of(md), getFmj(mods, md), imod));
                        }
                        else if (mods.toString().endsWith(".imod")) {
                            //TODO ensure this is not called while downloading a pack
                            //String fn = mods.getFileName().toString();
                            //if (!Files.exists(mods.getParent().resolve(fn.substring(0, fn.length() - 5))))
                            //    Files.delete(mods);
                            String fn = mods.getFileName().toString();
                            Path modFile = mods.getParent().resolve(fn.substring(0, fn.length() - 5));
                            ModDescription md = Utils.loadObject(mods, ModDescription.class);
                            evaluateSources(md, modFile);
                            descriptions.put(mods, new IWModDescription(mods, Optional.of(md), getFmj(modFile, md), Optional.of(mods)));
                        }
                        else {
                            descriptions.put(mods, new IWModDescription(mods));
                        }
                    }
                }
                Thread.sleep(100);
            } catch (IOException | URISyntaxException | InterruptedException e) {
                Inceptum.LOGGER.error("Could not list mods", e);
            }
        }
    }

    private void evaluateSources(ModDescription md, Path modFile) throws IOException {
        md.initialize(getGameVersion());
        boolean hasSource = false;
        for (ModSource source : md.sources.keySet()) {
            //Optional<ModSource> ms = source.getUpdate(instance.getMinecraftVersion());
            //TODO properly cache
            source.getUpdate(getGameVersion());
            if (!Files.exists(source.getJarPath())) source.download();
            hasSource = true;
            //if (ms.isEmpty()) continue;
            //if (update.isEmpty()) update = ms;
        }
        if (hasSource && Files.exists(modFile)) {
            Files.delete(modFile);
        }
    }

    private Optional<FabricModJson> getFmj(Path modJarDefault, ModDescription md) throws IOException, URISyntaxException {
        if (!Files.exists(modJarDefault)) {
            if (md.sources.isEmpty()) {
                throw new FileNotFoundException("Mod " + modJarDefault.getFileName().toString() + " doesn't specify a source and has no file");
            }
            modJarDefault = List.copyOf(md.sources.keySet()).get(0).getJarPath();
        }
        try (FileSystem fs = Utils.openZipFile(modJarDefault, false)) {
            return Files.exists(fs.getPath("fabric.mod.json"))
                    ? Optional.of(Utils.loadObject(fs.getPath("fabric.mod.json"), FabricModJson.class))
                    : Optional.empty();
        }
    }

    @Override
    public void close() {
        disposed = true;
    }

    public static record IWModDescription(Path path, Optional<ModDescription> mod, Optional<FabricModJson> fmj, Optional<Path> imod) implements Comparable<IWModDescription> {
        public String getName() {
            if (fmj.isEmpty()) return path.getFileName().toString();
            String base;
            if (fmj.get().name != null) base = fmj.get().name;
            else if (fmj.get().id != null) base = fmj.get().id;
            else {
                base = path.getFileName().toString();
                if (Files.isDirectory(path))
                    base += '/';
            }
            if (fmj.get().version != null) base += ' ' + fmj.get().version;
            return base;
        }

        public String[] getDescription() {
            try {
                if (fmj.isPresent() && fmj.get().description != null) {
                    return fmj.get().description.split("\n");
                }
                else if (Files.isDirectory(path)) {
                    return Utils.lsVi(path);
                }
            } catch (IOException e) {
                Inceptum.LOGGER.error("Could not get description", e);
            }
            return new String[] {"No description is available", "This mod may have been added manually"};
        }

        /**
         * Generates a description with only the path and no additional information
         * @param path The path of the mod entry
         */
        public IWModDescription(Path path) {
            this(path, Files.isDirectory(path) ? Optional.empty() : Optional.of(ModDescription.of(path)), Optional.empty(), Optional.empty());
        }

        /**
         * Generates a new mod description based on an existing one with an adjusted path
         * @param path The new path to use
         * @param base The mod description to copy
         */
        public IWModDescription(Path path, IWModDescription base) {
            this(path, base.mod, base.fmj, base.imod);
        }

        @Override
        public int compareTo(IWModDescription o) {
            return getName().compareTo(o.getName());
        }
    }
}
