package io.gitlab.jfronny.inceptum.windows;

import imgui.ImGui;
import imgui.type.ImBoolean;
import imgui.type.ImString;
import io.gitlab.jfronny.inceptum.Inceptum;
import io.gitlab.jfronny.inceptum.InceptumGui;
import io.gitlab.jfronny.inceptum.install.Steps;
import io.gitlab.jfronny.inceptum.model.inceptum.InstanceMeta;
import io.gitlab.jfronny.inceptum.model.inceptum.source.ModSource;
import io.gitlab.jfronny.inceptum.util.JvmUtils;
import io.gitlab.jfronny.inceptum.util.ModsDirScanner;
import io.gitlab.jfronny.inceptum.util.Utils;
import io.gitlab.jfronny.inceptum.windows.control.InstanceManageControls;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.Map;
import java.util.Optional;

public class InstanceEditWindow extends Window {
    private final Path path;
    private final InstanceMeta instance;
    private final InstanceManageControls imc;
    private final ImBoolean customJava;
    private final ImString customJavaPath = new ImString(128);
    private final ModsDirScanner mds;
    private Path selected = null;
    private boolean reDownload = false;

    public InstanceEditWindow(Path path, InstanceMeta instance) {
        super(path.getFileName().toString() + " - Edit");
        this.path = path;
        this.instance = instance;
        mds = new ModsDirScanner(path.resolve("mods"), instance);
        mds.start();
        customJava = new ImBoolean(instance.java != null);
        imc = new InstanceManageControls(instance);
    }

    @Override
    public void draw() {
        if (ImGui.beginTabBar("InstanceEdit" + path)) {
            if (ImGui.beginTabItem("General")) {
                if (ImGui.button("Open Directory")) {
                    Utils.openFile(path.toFile());
                }
                imc.nameBox("Rename", name -> {
                    try {
                        Path newPath = Inceptum.INSTANCE_DIR.resolve(name);
                        Files.move(path, newPath);
                        InceptumGui.open(new InstanceEditWindow(newPath, instance));
                        close();
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                });
                imc.snapshotsBox();
                imc.versionBox(ver -> {
                    reDownload = true;
                    instance.version = ver;
                    save();
                });
                if (ImGui.button("Delete")) Inceptum.showOkCancel("This instance will be removed forever (a long time)", "Are you sure?", () -> {
                    try {
                        Utils.deleteRecursive(path);
                    } catch (IOException e) {
                        Inceptum.showError("Could not delete the instance", e);
                    }
                    close();
                }, () -> {});
                if (ImGui.checkbox("Custom Java", customJava)) {
                    if (customJava.get()) {
                        instance.java = JvmUtils.getJvm();
                        customJavaPath.set(instance.java);
                    } else {
                        instance.java = null;
                    }
                    save();
                }
                if (customJava.get() && ImGui.inputText("Path", customJavaPath)) {
                    instance.java = customJavaPath.get();;
                    save();
                }
                ImGui.text("Did you know that every instance in Inceptum is a git repository?");
                ImGui.endTabItem();
            }
            //TODO update all/better updatability indicator
            //TODO there is still noticeable lag here
            //TODO drag-and-drop mods
            if (instance.isFabric() && ImGui.beginTabItem("Mods")) {
                if (!Files.exists(path.resolve("mods"))) {
                    try {
                        Files.createDirectories(path.resolve("mods"));
                    } catch (IOException e) {
                        Inceptum.LOGGER.error("Could not create mods directory which was missing from this modded instance", e);
                    }
                }
                ImGui.beginChild("mods select", 200, 0);
                if (ImGui.button("Add")) {
                    InceptumGui.WINDOWS.add(new AddModWindow(path.resolve("mods"), instance, mds));
                }
                ImGui.sameLine();
                if (Files.exists(path.resolve("mods")) && ImGui.button("Show")) {
                    Utils.openFile(path.resolve("mods").toFile());
                }
                ImGui.sameLine();
                if (Files.exists(path.resolve("config")) && ImGui.button("Configs")) {
                    Utils.openFile(path.resolve("config").toFile());
                }
                ImGui.separator();
                try {
                    for (ModsDirScanner.IWModDescription mod : mds.getMods()) {
                        boolean wasEnabled = !mod.path().toString().endsWith(".disabled");
                        if (ImGui.checkbox("##" + mod.getName(), wasEnabled)) {
                            String fName = mod.path().getFileName().toString();
                            final String disabledSuffix = ".disabled";
                            if (fName.endsWith(disabledSuffix))
                                fName = fName.substring(0, fName.length() - disabledSuffix.length());
                            if (wasEnabled) fName += disabledSuffix;
                            Path newSel = mod.path().getParent().resolve(fName);
                            try {
                                Files.move(mod.path(), newSel);
                                if (mod.path().equals(selected)) selected = newSel;
                                mod = new ModsDirScanner.IWModDescription(newSel, mod);
                            } catch (IOException e) {
                                Inceptum.showError("Could not change disabled state", e);
                            }
                        }
                        ImGui.sameLine();
                        if (ImGui.button(mod.getName())) selected = mod.path();
                    }
                } catch (IOException e) {
                    e.printStackTrace();
                }
                ImGui.endChild();
                ImGui.sameLine();
                ImGui.beginGroup();
                if (selected == null) {
                    ImGui.text("Select a mod to view settings");
                } else {
                    ModsDirScanner.IWModDescription md = mds.get(selected);
                    ImGui.text(md.getName());
                    ImGui.separator();
                    for (String s : md.getDescription()) {
                        ImGui.text(s);
                    }
                    ImGui.separator();
                    if (md.mod().isPresent()) {
                        Map<ModSource, Optional<ModSource>> sources = md.mod().get().sources;
                        ImGui.text("Sources:");
                        if (sources.isEmpty())
                            ImGui.bulletText("Local Drive");
                        else {
                            for (Map.Entry<ModSource, Optional<ModSource>> source : sources.entrySet()) {
                                ImGui.bulletText(source.getKey().getName());
                                Optional<ModSource> ms = source.getValue();
                                if (ms.isPresent()) {
                                    ImGui.sameLine();
                                    if (ImGui.button("Update to " + ms.get().getVersion())) {
                                        try {
                                            AddModWindow.DownloadMeta dm = AddModWindow.download(ms.get(), path.resolve("mods"), mds);
                                            dm.write();
                                            Files.delete(md.path());
                                            if (md.imod().isPresent() && Files.exists(md.imod().get()))
                                                Files.delete(md.imod().get());
                                            selected = dm.download().file();
                                        } catch (IOException e) {
                                            Inceptum.showError("Update failed", e);
                                        }
                                    }
                                }
                            }
                        }
                    }
                    if (ImGui.button("Delete")) {
                        if (md.mod().isPresent() && !md.mod().get().dependents.isEmpty())
                            Inceptum.showError("This mod still has the following dependent mods installed: " + String.join(", ", md.mod().get().dependents), "Dependents present");
                        else {
                            try {
                                delete(md);
                            } catch (IOException e) {
                                Inceptum.showError("Couldn't delete the file", e);
                            }
                            selected = null;
                        }
                    }
                }
                ImGui.endGroup();
                ImGui.endTabItem();
            }
            ImGui.endTabBar();
        }
    }

    private void delete(ModsDirScanner.IWModDescription md) throws IOException {
        Utils.deleteRecursive(md.path());
        if (md.imod().isPresent() && Files.exists(md.imod().get())) Files.delete(md.imod().get());
        if (md.mod().isPresent()) {
            for (String dependency : md.mod().get().dependencies) {
                Path dep = path.resolve("mods").resolve(dependency);
                ModsDirScanner.IWModDescription dmd = mds.get(dep);
                if (dmd.mod().isPresent()) {
                    dmd.mod().get().dependencies.remove(md.path().getFileName().toString());
                    if (dmd.mod().get().dependencies.isEmpty()) delete(dmd);
                    else if (dmd.imod().isPresent())
                        Utils.writeObject(dmd.imod().get(), dmd.mod().get());
                }
            }
        }
    }

    private void save() {
        try {
            Utils.writeObject(path.resolve("instance.json"), instance);
        } catch (IOException e) {
            Inceptum.LOGGER.error("Could not write instance config", e);
        }
    }

    @Override
    public void close() {
        super.close();
        mds.close();
        if (reDownload) {
            try {
                //TODO show progress
                Steps.reDownload(path);
            } catch (IOException e) {
                Inceptum.showError("Could not re-download data", e);
            }
        }
    }
}
