package io.gitlab.jfronny.inceptum.windows;

import com.sun.net.httpserver.HttpServer;
import imgui.ImGui;
import io.gitlab.jfronny.inceptum.Inceptum;
import io.gitlab.jfronny.inceptum.model.microsoft.*;
import io.gitlab.jfronny.inceptum.util.Utils;
import io.gitlab.jfronny.inceptum.util.api.account.AccountManager;
import io.gitlab.jfronny.inceptum.util.api.account.MicrosoftAccount;
import io.gitlab.jfronny.inceptum.util.api.account.MicrosoftAuthAPI;

import java.io.OutputStream;
import java.net.InetSocketAddress;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URLDecoder;
import java.nio.charset.StandardCharsets;
import java.util.LinkedHashMap;
import java.util.Map;

public class MicrosoftLoginWindow extends Window {
    private static HttpServer server;
    private static final String MICROSOFT_LOGIN_URL = "https://login.live.com/oauth20_authorize.srf?client_id=90890812-00d1-48a8-8d3f-38465ef43b58&prompt=select_account&cobrandid=8058f65d-ce06-4c30-9559-473c9275a65d&response_type=code&scope=XboxLive.signin%20XboxLive.offline_access&redirect_uri=http%3A%2F%2F127.0.0.1%3A28562";

    private final MicrosoftAccount account;

    public MicrosoftLoginWindow() {
        this(null);
    }

    public MicrosoftLoginWindow(MicrosoftAccount account) {
        super("Microsoft Login");

        this.account = account;
        try {
            startServer();
        } catch (Exception e) {
            Inceptum.LOGGER.error("Could not start mc login server", e);
        }
    }

    @Override
    public void draw() {
        ImGui.text("This feature uses modified ATLauncher code, so the login prompt will ask you to log in to ATLauncher");
        ImGui.text("Click the button below to begin");
        if (ImGui.button("Open in Browser")) {
            try {
                Utils.openWebBrowser(new URI(MICROSOFT_LOGIN_URL));
            } catch (URISyntaxException e) {
                Inceptum.LOGGER.error("Could not open browser", e);
            }
        }
    }

    private void startServer() throws Exception {
        if (server != null) return;
        server = HttpServer.create(new InetSocketAddress("127.0.0.1", MicrosoftAuthAPI.MICROSOFT_LOGIN_REDIRECT_PORT), 0);

        server.createContext("/", req -> {
            if (req.getRequestMethod().equals("GET")) {
                Map<String, String> query = new LinkedHashMap<>();
                for (String pair : req.getRequestURI().getQuery().split("&")) {
                    int index = pair.indexOf('=');
                    query.put(URLDecoder.decode(pair.substring(0, index), StandardCharsets.UTF_8), URLDecoder.decode(pair.substring(index + 1), StandardCharsets.UTF_8));
                }

                int respCode;
                String respStr;

                if (query.containsKey("error")) {
                    respCode = 500;
                    respStr = "Error logging in. Check console for more information";
                    Inceptum.LOGGER.error("Error logging into Microsoft account: " + URLDecoder
                            .decode(query.get("error_description"), StandardCharsets.UTF_8.toString()));
                }
                else if (query.containsKey("code")) {
                    try {
                        acquireAccessToken(query.get("code"));
                        respCode = 200;
                        respStr = "Login complete. You can now close this window and go back to Inceptum";
                    } catch (Exception e) {
                        Inceptum.LOGGER.error("Error acquiring accessToken", e);
                        respCode = 500;
                        respStr = "Error logging in. Check console for more information";
                    }
                }
                else {
                    respCode = 400;
                    respStr = "Code is missing";
                }

                OutputStream os = req.getResponseBody();
                req.sendResponseHeaders(respCode, respStr.length());
                os.write(respStr.getBytes(StandardCharsets.UTF_8));
                os.flush();
                os.close();

                stopServer();
            }
        });

        server.start();
    }

    private void stopServer() {
        if (server == null) return;
        server.stop(0);
        server = null;
    }

    private void acquireAccessToken(String authcode) throws Exception {
        OauthTokenResponse oauthTokenResponse = MicrosoftAuthAPI.tradeCodeForAccessToken(authcode);

        acquireXBLToken(oauthTokenResponse);
    }

    private void acquireXBLToken(OauthTokenResponse oauthTokenResponse) throws Exception {
        XboxLiveAuthResponse xblAuthResponse = MicrosoftAuthAPI.getXBLToken(oauthTokenResponse.accessToken);

        acquireXsts(oauthTokenResponse, xblAuthResponse.token);
    }

    private void acquireXsts(OauthTokenResponse oauthTokenResponse, String xblToken) throws Exception {
        XboxLiveAuthResponse xstsAuthResponse = MicrosoftAuthAPI.getXstsToken(xblToken);

        if (xstsAuthResponse != null) {
            acquireMinecraftToken(oauthTokenResponse, xstsAuthResponse);
        }
    }

    private void acquireMinecraftToken(OauthTokenResponse oauthTokenResponse, XboxLiveAuthResponse xstsAuthResponse) throws Exception {
        String xblUhs = xstsAuthResponse.displayClaims.xui.get(0).uhs;
        String xblXsts = xstsAuthResponse.token;

        LoginResponse loginResponse = MicrosoftAuthAPI.loginToMinecraft("XBL3.0 x=" + xblUhs + ";" + xblXsts);

        Store store = MicrosoftAuthAPI.getMcEntitlements(loginResponse.accessToken);

        Inceptum.LOGGER.info(Inceptum.GSON.toJson(store));

        if (!(store.items.stream().anyMatch(i -> i.name.equalsIgnoreCase("product_minecraft"))
                && store.items.stream().anyMatch(i -> i.name.equalsIgnoreCase("game_minecraft")))) {
            Inceptum.showError("This account doesn't have a valid purchase of Minecraft.\nPlease make sure you've bought the Java edition of Minecraft and then try again.", "Doesn't own Minecraft");
            throw new Exception("Account does not own Minecraft");
        }

        Profile profile = MicrosoftAuthAPI.getMcProfile(loginResponse.accessToken);

        if (profile == null) {
            throw new Exception("Failed to get Minecraft profile");
        }

        // add the account
        addAccount(oauthTokenResponse, xstsAuthResponse, loginResponse, profile);

        close();
        Inceptum.showInfo("The account \"" + profile.name + "\" was added successfully", "Account added");
    }

    private void addAccount(OauthTokenResponse oauthTokenResponse, XboxLiveAuthResponse xstsAuthResponse,
                            LoginResponse loginResponse, Profile profile) throws Exception {
        if (account != null || AccountManager.isAccountByName(loginResponse.username)) {
            MicrosoftAccount account = (MicrosoftAccount) AccountManager.getAccountByName(loginResponse.username);

            if (account == null) {
                return;
            }

            // if forced to relogin, then make sure they logged into correct account
            if (this.account != null && account.accountId != this.account.accountId) {
                Inceptum.showError("Logged into incorrect account. Please login again on the Accounts tab", "Incorrect account");
                return;
            }

            account.update(oauthTokenResponse, xstsAuthResponse, loginResponse, profile);
            AccountManager.saveAccounts();
        } else {
            MicrosoftAccount account = new MicrosoftAccount(oauthTokenResponse, xstsAuthResponse, loginResponse,
                    profile);

            AccountManager.addAccount(account);
        }
    }
}
