package io.gitlab.jfronny.inceptum;

import io.gitlab.jfronny.inceptum.model.ComparableVersion;
import io.gitlab.jfronny.inceptum.model.inceptum.UpdateChannel;
import io.gitlab.jfronny.inceptum.model.inceptum.UpdateInfo;
import io.gitlab.jfronny.inceptum.util.JvmUtils;
import io.gitlab.jfronny.inceptum.util.MetaHolder;
import io.gitlab.jfronny.inceptum.util.OSCheck;
import io.gitlab.jfronny.inceptum.util.UpdateChecker;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.stream.Stream;

public class Wrapper {
    private static final Path INCEPTUM_LIB_DIR = MetaHolder.BASE_PATH.resolve("libraries/io/gitlab/jfronny/inceptum/Inceptum");

    public static void main(String[] args) throws IOException {
        System.out.println("Inceptum Wrapper v" + MetaHolder.VERSION.version + " (" + MetaHolder.VERSION.flavor + ")");
        if (!Files.exists(INCEPTUM_LIB_DIR)) downloadLatest();
        Optional<Path> localBinary = selectLocalBinary();
        if (localBinary.isEmpty()) {
            downloadLatest();
            localBinary = selectLocalBinary();
        }
        if (localBinary.isEmpty()) {
            throw new FileNotFoundException("Could not identify or download a valid binary");
        }
        List<String> newArgs = new ArrayList<>();
        newArgs.add(JvmUtils.getJvm());
        newArgs.add("-jar");
        newArgs.add(localBinary.get().toString());
        newArgs.add("wrapper");
        newArgs.addAll(Arrays.asList(args));
        new ProcessBuilder(newArgs.toArray(new String[0]))
                .inheritIO()
                .start();
    }

    private static void downloadLatest() throws IOException {
        System.err.println("No Inceptum jar was identified");
        UpdateInfo ui = UpdateChecker.check(UpdateChannel.Stable, new ComparableVersion("0"), OSCheck.OS.getMojName(), c -> {}, System.out::println, System.err::println, (s, e) -> {
            e.printStackTrace();
            System.err.println(s);
        });
        if (ui == null) {
            throw new FileNotFoundException("Could not identify a valid inceptum version. Are you connected to the internet?");
        }
        System.err.println("Downloading " + ui.url());
        Path dir = INCEPTUM_LIB_DIR.resolve(ui.newVersion().toString());
        Files.createDirectories(dir);
        try (InputStream is = new URL(ui.url()).openStream()) {
            Files.write(dir.resolve("Inceptum-" + ui.newVersion() + '-' + OSCheck.OS.getMojName() + ".jar"), is.readAllBytes());
        }
        catch (Throwable t) {
            Files.delete(dir);
        }
    }

    private static Optional<Path> selectLocalBinary() throws IOException {
        Path pathChosen = null;
        ComparableVersion versionChosen = new ComparableVersion("0");
        try (Stream<Path> inceptumLibVersionDirStream = Files.list(INCEPTUM_LIB_DIR)) {
            for (Path inceptumVersionPath : inceptumLibVersionDirStream.toList()) {
                ComparableVersion versionCurrent = new ComparableVersion(inceptumVersionPath.getFileName().toString());
                if (versionCurrent.compareTo(versionChosen) > 0) {
                    inceptumVersionPath = inceptumVersionPath.resolve("Inceptum-" + versionCurrent + '-' + OSCheck.OS.getMojName() + ".jar");
                    if (Files.exists(inceptumVersionPath)) {
                        versionChosen = versionCurrent;
                        pathChosen = inceptumVersionPath;
                    }
                    else {
                        System.out.println("Candidate " + inceptumVersionPath + " failed: doesn't exist");
                    }
                }
            }
        }
        if (pathChosen == null) {
            System.err.println("Something went wrong, please try again");
            return Optional.empty();
        }
        return Optional.of(pathChosen);
    }
}
